@@ -31,13 +31,19 @@ generate the model of a tree on disk, or a new journal with redundant changes
 eliminated.
 
 DiskUpdater is a class to compare a memory 'tree' and a bzr transport and 
-output a journal to update the tree to match the transport.
+output a journal to update the tree to match the transport. DiskUpdater uses
+a helper filter object to do path filtering. FilterCombiner and ProcessFilter
+are useful helpers for such filtering.
 
 Various Replay objects can replay a journal. TransportReplay replays a journal
-by reading the file content from a transport object.
+by reading the file content from a transport object. ReplayGenerator is the
+core workhorse for streaming the data needed to do a replay; it is serialised
+when streaming from http and deserialised by FromFileGenerator.
 """
 
-__all__ = ['parse', 'Combiner', 'Journal', 'DiskUpdater', 'TransportReplay']
+__all__ = ['parse', 'Combiner', 'Journal', 'DiskUpdater', 'TransportReplay',
+    'FilterCombiner', 'ProcessFilter',
+    ]
 
 import errno
 import os
@@ -240,10 +246,17 @@ class DiskUpdater(object):
         '^foo/(?!bar(?$|/)', or exclude paths starting with foo, and include
         both foo and bar explicitly: exclude '^foo(?:$|/)' and include '^foo$',
         '^foo/bar(?$|/)'.
+    :ivar filter_callback: A filter which is applid to all paths. If the filter
+        returns False for a path, the path is considered to be excluded (as if
+        it had matched the exclude_re); if the filter returns True, then the
+        path is considered included as if it matched the include_re. Finally,
+        if the filter returns None, the path is not influenced by the filter
+        callback.
     """
 
     def __init__(self, tree, transport, name, last_timestamp, ui,
-        excludes=(), includes=()):
+        excludes=(), includes=(), filter_callback=lambda path:None,
+        known_changes=None):
         """Create a DiskUpdater.
 
         :param tree: The tree to compare with.
@@ -259,6 +272,14 @@ class DiskUpdater(object):
             the exclude_re.
         :param includes: An optional list of uncompiled regexes to include in
             the include_re.
+        :param filter_callback: A path filter. See the class docstring for
+            details.
+        :param known_changes: Either None, or a list of abspaths for changes
+            known within the content root being scanned. If not None, then
+            the directory structure on disk is not walked; rather the known
+            changes are used to detect changes. As a special case, the children
+            of directories within known_changes that were not present in the
+            last mirror update are scanned on disk.
         """
         self.tree = tree
         self.transport = transport
@@ -271,17 +292,61 @@ class DiskUpdater(object):
         self.include_re = re.compile(self._make_re_str(includes))
         excludes = [r'(?:^|/)\.lmirror/'] + list(excludes)
         self.exclude_re = re.compile(self._make_re_str(excludes))
+        self.filter_callback = filter_callback
+        self.known_changes = known_changes
 
     def _make_re_str(self, re_strs):
         re_strs = ['(?:%s)' % re_str for re_str in re_strs]
         return '|'.join(re_strs)
 
+    def _real_dir_contents(self, dirname):
+        """Get the contents of dirname from disk."""
+        return self.transport.list_dir(dirname)
+    
+    def _known_dir_contents(self, dirname):
+        """Get the contents of dirname from self.known_changes."""
+        cwd = self._known_changes_dirs
+        if dirname == '':
+            return cwd.keys()
+        elements = dirname.split('/')
+        for element in elements:
+            cwd = cwd[element][1]
+        return cwd.keys()
+
     def finished(self):
         """Return the journal obtained by scanning the disk."""
+        if self.known_changes is None:
+            return self._finished_scan(self._real_dir_contents)
+        else:
+            self._cache_known_changes()
+            return self._finished_scan(self._known_dir_contents,
+                missing_is_unchanged=True)
+
+    def _cache_known_changes(self):
+        """Structure known_changes into a dict for lookups."""
+        root_prefix = self.transport.local_abspath('.')
+        root = {}
+        for path in self.known_changes:
+            path = path[len(root_prefix) + 1:]
+            elements = path.split('/')
+            cwd = root
+            for pos, element in enumerate(elements):
+                dirname = '/'.join(elements[:pos + 1])
+                subdir = cwd.setdefault(element, (dirname, {}))
+                cwd = subdir[1]
+        self._known_changes_dirs = root
+
+    def _finished_scan(self, dir_contents, missing_is_unchanged=False):
+        """Perform finished() by scanning the disk.
+        
+        :param dir_contents: A callback to get the contents of a directory.
+        :param missing_is_unchanged: If True, a path not listed in a directory
+            is unchanged, rather than missing.
+        """
         pending = ['']
         while pending:
             dirname = pending.pop(-1)
-            names = self.transport.list_dir(dirname)
+            names = dir_contents(dirname)
             # NB: quadratic in lookup here due to presence in inner loop:
             # consider tuning.
             segments = dirname.split('/')
@@ -295,28 +360,53 @@ class DiskUpdater(object):
                     # totally new directory - added to journal by the directory
                     # above.
                     cwd = {}
+            # tree_names contains the last recorded set of names.
             tree_names = set(cwd)
             names = set(names)
-            for name in tree_names - names:
-                # deletes
-                path = dirname and ('%s/%s' % (dirname, name)) or name
-                old_kind_details = cwd[name]
-                if type(old_kind_details) is dict:
-                    self._gather_deleted_dir(path, cwd[name])
-                    old_kind_details = ('dir',)
-                self.journal.add(path, 'del',
-                    old_kind_details)
+            if not missing_is_unchanged:
+                for name in tree_names - names:
+                    # deletes
+                    path = dirname and ('%s/%s' % (dirname, name)) or name
+                    old_kind_details = cwd[name]
+                    if type(old_kind_details) is dict:
+                        self._gather_deleted_dir(path, old_kind_details)
+                        old_kind_details = DirContent()
+                    self.journal.add(path, 'del', old_kind_details)
             new_names = names - tree_names
             for name in names:
                 path = dirname and ('%s/%s' % (dirname, name)) or name
                 if self._skip_path(path):
+                    if name in tree_names:
+                        # Newly excluded.
+                        old_kind_details = cwd[name]
+                        if type(old_kind_details) is dict:
+                            self._gather_deleted_dir(path, old_kind_details)
+                            old_kind_details = DirContent()
+                        self.journal.add(path, 'del', old_kind_details)
+                    continue
+                try:
+                    statinfo = self.transport.stat(path)
+                except errors.NoSuchFile:
+                    # This file doesn't actually exist: may be concurrent
+                    # delete, or a seen change from a changes list.
+                    if name in tree_names:
+                        # A delete.
+                        old_kind_details = cwd[name]
+                        if type(old_kind_details) is dict:
+                            self._gather_deleted_dir(path, old_kind_details)
+                            old_kind_details = DirContent()
+                        self.journal.add(path, 'del', old_kind_details)
                     continue
-                statinfo = self.transport.stat(path)
-                # Is it old enough to not check
                 mtime = getattr(statinfo, 'st_mtime', 0)
-                if self.last_timestamp - mtime > 3 and name not in new_names:
-                    continue
                 kind = osutils.file_kind_from_stat_mode(statinfo.st_mode)
+                if (kind != 'directory' and self.last_timestamp - mtime > 3
+                    and name not in new_names):
+                    # We have to look inside directories always; things that
+                    # are older than 3 seconds we can trust even FAT to not
+                    # be lying about the last-modification (it has 2 second
+                    # granularity) and finally its not new (new things have
+                    # to be scanned always).
+                    continue
                 if kind == 'file':
                     f = self.transport.get(path)
                     try:
@@ -350,11 +440,11 @@ class DiskUpdater(object):
         pending = [(path, dirdict)]
         while pending:
             dirname, cwd = pending.pop(-1)
-            for name, old_kind_details in cwd:
+            for name, old_kind_details in cwd.iteritems():
                 path = dirname and ('%s/%s' % (dirname, name)) or name
                 if type(old_kind_details) is dict:
                     pending.append((path, old_kind_details))
-                    old_kind_details = ('dir',)
+                    old_kind_details = DirContent()
                 self.journal.add(path, 'del', old_kind_details)
 
     def _skip_path(self, path):
@@ -363,10 +453,94 @@ class DiskUpdater(object):
             path.endswith('.lmirrortemp')):
             # metadata is transmitted by the act of fetching the
             # journal.
+            self.ui.output_log(1, __name__,
+                "Skipping %r because it is lmirror metadata/temp file" % path)
             return True
-        if self.exclude_re.search(path) and not self.include_re.search(path):
-            return True
-        return False
+        filter_result = self.filter_callback(path)
+        excluded = filter_result is False
+        included = filter_result is True
+        if excluded or self.exclude_re.search(path):
+            if not (included or self.include_re.search(path)):
+                self.ui.output_log(1, __name__,
+                    "Skipping %r because it is excluded." % path)
+                return True
+            else:
+                self.ui.output_log(1, __name__,
+                    "Included %r because it is included." % path)
+                return False
+        else:
+            self.ui.output_log(1, __name__,
+                "Included %r because it is not excluded." % path)
+            return False
+
+
+class FilterCombiner(object):
+    """An updater filter to combine other filters.
+
+    This holds a list of filters to combine.  Filters that return True short
+    circuit further evaluation. If no filter returns True, if any return False
+    False is returned, otherwise None.
+
+    :ivar filters: The filters.
+    """
+
+    def __init__(self, *filters):
+        """Create a FilterCombiner.
+
+        :param filters: The filters to combine.
+        """
+        self.filters = filters
+
+    def __call__(self, path):
+        result = None
+        for filter in self.filters:
+            next_result = filter(path)
+            if next_result:
+                return True
+            if next_result is False:
+                result = next_result
+        return result
+
+
+class ProcessFilter(object):
+    """A filter that uses an external process to perform filtering.
+
+    The process is communicated with via a line based protocol. For each path
+    to be filtered, the path + \n are written to the process, and a response
+    read back in. The response should be one of True\n, False\n or None\n.
+
+    :ivar proc: The process being used to do the filtering. This must be a 
+        subprocess.Popen or similar object; in particular its stdin must
+        support write(), and its stdout must support readline(). The process
+        is not started, not closed by ProcessFilter - the caller should take
+        care of that.
+    :ivar ui: The UI for logging.
+    :ivar description: The description of the process for logging.
+    """
+
+    def __init__(self, proc, ui, description):
+        """Create a ProcessFilter using proc to do the filtering.
+
+        :param proc: A subprocess.Popen or similar object with stdin and stdout
+            file like objects that can have write() and readline() called on them.
+        :param ui: UI to log actions to.
+        :param description: How to describe the helper.
+        """
+        self.proc = proc
+        self._results = {'True\n': True, 'False\n': False, 'None\n': None}
+        self.ui = ui
+        self.description = description
+
+    def __call__(self, path):
+        """Filter path. See FilterCombiner's docstring for details.
+        
+        :param path: The path to filter.
+        """
+        self.proc.stdin.write("%s\n" % path)
+        result = self._results[self.proc.stdout.readline()]
+        self.ui.output_log(1, __name__, "helper %s filtered %r with result %r"
+            % (self.description, path, result))
+        return result
 
 
 class Journal(object):
@@ -526,6 +700,9 @@ class Action(object):
         self.path = path
         self.content = content
 
+    def __repr__(self):
+        return "Action: %r %s %s" % (self.path, self.type, self.content)
+
     def as_bytes(self):
         """Return a generator of bytes for this action.
 
@@ -571,21 +748,27 @@ class TransportAction(Action):
     :ivar sourcedir: Transport to read file content from.
     """
 
-    def __init__(self, action_type, path, content, sourcedir):
+    def __init__(self, action_type, path, content, sourcedir, ui):
         Action.__init__(self, action_type, path, content)
         self.sourcedir = sourcedir
+        self.ui = ui
 
     def get_file(self):
         """Get the content for a new file as a file-like object."""
         return self.sourcedir.get(self.path)
 
+    def ignore_file(self):
+        self.ui.output_log(
+            4, __name__, 'Ignoring %s %r' % (self.content.kind, self.path))
+
 
 class StreamedAction(Action):
     """An Action which gets file content from a FromFileGenerator."""
 
-    def __init__(self, action_type, path, content, generator):
+    def __init__(self, action_type, path, content, generator, ui):
         Action.__init__(self, action_type, path, content)
         self.generator = generator
+        self.ui = ui
 
     def get_file(self):
         if self.type == 'replace':
@@ -595,23 +778,27 @@ class StreamedAction(Action):
         if content.kind != 'file':
             raise ValueError('invalid call to get_file: kind is %r' %
                 content.kind)
-        return BufferedFile(self.generator, content.length)
+        return BufferedFile(self.generator, content.length, self.ui)
 
     def ignore_file(self):
+        self.ui.output_log(
+            4, __name__, 'Ignoring %s %r' % (self.content.kind, self.path))
         self.get_file().close()
 
 
 class BufferedFile(object):
     """A file-like object which reads from a FromFileGenerator's stream."""
 
-    def __init__(self, generator, remaining):
+    def __init__(self, generator, remaining, ui):
         self.generator = generator
         self.remaining = remaining
+        self.ui = ui
 
     def read(self, count=None):
         if count is None:
             count = self.remaining
         read_size = min(self.remaining, count)
+        self.ui.output_log(1, __name__, "Reading from stream, read_size=%d remaining=%d" % (read_size, self.remaining))
         if not read_size:
             return ''
         read_content = self.generator._next_bytes(read_size)
@@ -650,7 +837,8 @@ class ReplayGenerator(object):
         groups = self.journal.as_groups()
         for group in groups:
             for action, path, content in group:
-                yield TransportAction(action, path, content, self.sourcedir)
+                yield TransportAction(
+                    action, path, content, self.sourcedir, self.ui)
 
     def as_bytes(self):
         """Return a generator of bytestrings for this generator's content."""
@@ -665,9 +853,10 @@ class FromFileGenerator(object):
     This is used for streaming from HTTP servers.
     """
 
-    def __init__(self, stream):
+    def __init__(self, stream, ui):
         self._stream = stream
         self.buffered_bytes = []
+        self.ui = ui
 
     def parse_kind_data(self, tokens):
         pos = 2
@@ -701,7 +890,7 @@ class FromFileGenerator(object):
             else:
                 raise ValueError('unknown action %r' % action)
             self._push('\x00'.join(tokens[pos:]))
-            yield StreamedAction(action, path, kind_data, self)
+            yield StreamedAction(action, path, kind_data, self, self.ui)
 
     def _next_bytes(self, count):
         """Return up to count bytes.
@@ -714,7 +903,10 @@ class FromFileGenerator(object):
             return self._stream.read(count)
         if count >= len(self.buffered_bytes[0]):
             partial = self.buffered_bytes.pop(0)
-            return partial + self._next_bytes(count - len(partial))
+            if len(partial) < count:
+                return partial + self._next_bytes(count - len(partial))
+            else:
+                return partial
         result = self.buffered_bytes[0][:count]
         self.buffered_bytes[0] = self.buffered_bytes[0][count:]
         return result
@@ -759,12 +951,15 @@ class TransportReplay(object):
     def replay(self):
         """Replay the journal."""
         groups = self.journal.as_groups()
-        for group in groups:
+        for pos, group in enumerate(groups):
+            self.ui.output_log(4, __name__, "Processing group %d of %d with %d elements" % (pos, len(groups), len(group)))
             elements = set(group)
+            assert len(elements) == len(group)
             to_rename = []
             to_delete = []
             try:
                 while elements:
+                    self.ui.output_log(3, __name__, "Waiting for element, %d remaining" % (len(elements),))
                     action_obj = self.generator.next()
                     # If this fails, generator has sent us some garbage.
                     elements.remove((action_obj.type, action_obj.path,
@@ -826,6 +1021,7 @@ class TransportReplay(object):
                 raise ValueError('unexpected non-file at %r' % path)
             f = self.contentdir.get(path)
             try:
+                self.ui.output_log(4, __name__, 'Hashing %s %r' % (content.kind, path))
                 size, sha1 = osutils.size_sha_file(f)
             finally:
                 f.close()
@@ -838,10 +1034,8 @@ class TransportReplay(object):
 
         :param tempname: The name of a temporary file with the needed content.
         """
-        if self.check_file(path, content):
-            # Note that this implies we copied a file we didn't - during
-            # a complete resync - sure, but still not optimal bw use.
-            self.contentdir.delete(tempname)
+        if self.contentdir.has(path):
+            self.contentdir.delete(path)
         self.contentdir.rename(tempname, path)
 
     def ensure_link(self, realpath, target):
@@ -850,14 +1044,14 @@ class TransportReplay(object):
         An error is raised if something is in the way.
         """
         try:
-            os.symlink(content[1], realpath)
-        except OSerror, e:
+            os.symlink(target, realpath)
+        except OSError, e:
             if e.errno == errno.EEXIST:
                 st = os.lstat(realpath)
                 if osutils.file_kind_from_stat_mode(st.st_mode) != 'symlink':
                     raise ValueError('unexpected non-symlink at %r' % realpath)
                 os.unlink(realpath)
-                os.symlink(content[1], realpath)
+                os.symlink(target, realpath)
             else:
                 raise
 
@@ -871,7 +1065,7 @@ class TransportReplay(object):
             IO has been done before returning.
         """
         tempname = '%s.lmirrortemp' % path
-        self.ui.output_log(4, __name__, 'Writing %s %r' % (content.kind, path))
+        self.ui.output_log(4, __name__, 'Checking %s %r' % (content.kind, path))
         if content.kind == 'dir':
             return lambda: self.ensure_dir(path)
         elif content.kind == 'symlink':
@@ -884,7 +1078,9 @@ class TransportReplay(object):
             if self.check_file(path, content):
                 action.ignore_file()
                 return lambda:None
-        except ValueError:
+        except (ValueError, IOError):
+            # If we can't read the file for some reason, we obviously need to
+            # write it :).
             pass
         a_file = action.get_file()
         source = _ShaFile(a_file)
