@@ -29,11 +29,14 @@ working with a MirrorSet.
 __all__ = ['initialise', 'MirrorSet']
 
 import ConfigParser
+import json
 from StringIO import StringIO
+import subprocess
 import time
 
 from bzrlib import urlutils
 from bzrlib.errors import NoSuchFile, NotLocalUrl
+from bzrlib.transport import get_transport
 
 from l_mirror import gpg, journals
 
@@ -121,6 +124,9 @@ class _MirrorSet(object):
         apply when scanning for changes.
     :ivar includes: () if not loaded from disk, or the include regexes to
         apply when scanning for changes.
+    :ivar filter_programs: () if not loaded from disk, or the list of programs
+        and arguments to run and use when scanning for changes in this
+        mirrorset.
     :ivar ui: The AbstractUI output is fed to.
     :ivar gpg_strategy: A bzrlib.gpg.GPGStrategy used for doing gpg signatures.
     :ivar gpgv_strategy: A l_mirror.gpg.GPGVStrategy for doing signature
@@ -140,6 +146,7 @@ class _MirrorSet(object):
         self.ui = ui
         self.excludes = ()
         self.includes = ()
+        self.filter_programs = ()
         self.gpg_strategy = gpg.SimpleGPGStrategy(None)
         try:
             self.gpgv_strategy = gpg.GPGVStrategy(
@@ -148,10 +155,13 @@ class _MirrorSet(object):
             # won't be able to do gpgv calls.
             self.gpgv_strategy = None
 
-    def finish_change(self):
+    def finish_change(self, dryrun=False):
         """Scan the mirror set for changes and write a new journal entry.
 
         This will set updating=False and update the timestamp in the metadata.
+
+        :param dryrun: If True perform the content scan (and log/report it as
+            appropriate) but do not change the metadata or write a new journal.
         """
         metadata = self._get_metadata()
         if metadata.get('metadata', 'updating') != 'True':
@@ -163,25 +173,49 @@ class _MirrorSet(object):
         now = time.time()
         basis = int(metadata.get('metadata', 'basis'))
         latest = int(metadata.get('metadata', 'latest'))
-        current_state = self._combine_journals(basis, latest)
-        updater = journals.DiskUpdater(current_state,
-            self._content_root_dir(), self.name, last, self.ui,
-            includes = self.get_includes(), excludes=self.get_excludes())
-        journal = updater.finished()
-        if journal.paths:
-            next_id = latest + 1
-            journal_bytes = journal.as_bytes()
-            journal_dir = self._journaldir()
-            journal_dir.put_bytes(str(next_id), journal_bytes)
-            if self._is_signed():
-                signature = self.gpg_strategy.sign(journal_bytes)
-                journal_dir.put_bytes("%s.sig" % next_id, signature)
-            metadata.set('metadata', 'latest', str(next_id))
-            metadata.set('metadata', 'timestamp', str(now))
+        if metadata.has_option('metadata', 'server'):
+            server_url = metadata.get('metadata', 'server')
+            server_transport = get_transport(server_url)
+            try:
+                changes_bytes = server_transport.get_bytes('changes/%s' % self.name)
+                # TODO: May require URL decoding to match the transport interface.
+                changes = sorted(path.encode('utf8') for path in
+                    json.loads(changes_bytes))
+            except NoSuchFile:
+                server_transport = None
+                changes = None
         else:
-            self.ui.output_rest('No changes found in mirrorset.')
-        metadata.set('metadata', 'updating', 'False')
-        self._set_metadata(metadata)
+            server_transport = None
+            changes = None
+        current_state = self._combine_journals(basis, latest)
+        filter_callback = self._get_filter_callback()
+        try:
+            updater = journals.DiskUpdater(current_state,
+                self._content_root_dir(), self.name, last, self.ui,
+                includes = self.get_includes(), excludes=self.get_excludes(),
+                filter_callback=filter_callback, known_changes=changes)
+            journal = updater.finished()
+            if not dryrun and journal.paths:
+                next_id = latest + 1
+                journal_bytes = journal.as_bytes()
+                journal_dir = self._journaldir()
+                journal_dir.put_bytes(str(next_id), journal_bytes)
+                if self._is_signed():
+                    signature = self.gpg_strategy.sign(journal_bytes)
+                    journal_dir.put_bytes("%s.sig" % next_id, signature)
+                metadata.set('metadata', 'latest', str(next_id))
+                metadata.set('metadata', 'timestamp', str(now))
+            else:
+                self.ui.output_rest('No changes found in mirrorset.')
+            metadata.set('metadata', 'updating', 'False')
+            if not dryrun:
+                self._set_metadata(metadata)
+                if server_transport is not None:
+                    server_transport.get_bytes('updated/%s' % self.name)
+        finally:
+            for filter in filter_callback.filters:
+                # Signal it should close and wait for it.
+                filter.proc.communicate('')
 
     def get_excludes(self):
         if self.excludes == ():
@@ -193,6 +227,23 @@ class _MirrorSet(object):
             self._parse_content_conf()
         return self.includes
 
+    def _get_filter_callback(self):
+        if self.filter_programs == ():
+            self._parse_content_conf()
+        filters = []
+        for program in self.filter_programs:
+            proc = self.ui.subprocess_Popen(program, stdin=subprocess.PIPE,
+                stdout=subprocess.PIPE)
+            filters.append(journals.ProcessFilter(proc, self.ui, program))
+        return journals.FilterCombiner(*filters)
+
+    def get_server(self):
+        """Get the current server for this node from metadata.conf."""
+        metadata = self._get_metadata()
+        if not metadata.has_option('metadata', 'server'):
+            return None
+        return metadata.get('metadata', 'server')
+
     def _contentdir(self):
         """Return a transport rooted at the content of this mirror set."""
         return self.base
@@ -329,10 +380,23 @@ class _MirrorSet(object):
     def _setdir(self):
         return self.base.clone('.lmirror/sets/%s' % self.name)
 
+    def set_server(self, server):
+        """Set a current server in metadata.conf.
+
+        :param server: The server URL to remember.
+        """
+        metadata = self._get_metadata()
+        if server is None:
+            metadata.remove_option('metadata', 'server')
+        else:
+            metadata.set('metadata', 'server', server)
+        self._set_metadata(metadata)
+
     def _parse_content_conf(self):
         t = self._setdir()
         includes = []
         excludes = []
+        programs = []
         try:
             file_bytes = t.get_bytes('content.conf')
             for line in file_bytes.split('\n'):
@@ -342,10 +406,13 @@ class _MirrorSet(object):
                     includes.append(line[8:])
                 elif line.startswith('exclude '):
                     excludes.append(line[8:])
+                elif line.startswith('program '):
+                    programs.append(line[8:])
         except NoSuchFile:
             pass
         self.includes = includes
         self.excludes = excludes
+        self.filter_programs = programs
 
     def content_root_path(self):
         return self._get_settings().get('set', 'content_root')
@@ -398,7 +465,7 @@ class HTTPMirrorSet(_MirrorSet):
         # Work around https://bugs.edge.launchpad.net/bzr/+bug/555032
         code, stream = self.base._get('stream/%s/%s/%s' % (self.name, from_journal,
             to_journal), None)
-        return journals.FromFileGenerator(stream)
+        return journals.FromFileGenerator(stream, self.ui)
 
 
 class OrderedConfigParser(ConfigParser.ConfigParser):
