@@ -163,6 +163,17 @@ exclude another regex
         # sufficient.
         mirror.finish_change()
 
+    def test_set_server(self):
+        basedir = get_transport(self.setup_memory()).clone('path')
+        basedir.create_prefix()
+        ui = self.get_test_ui()
+        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
+        mirror.finish_change()
+        mirror.set_server('foo/bar')
+        self.assertEqual('foo/bar', mirror.get_server())
+        mirror.set_server(None)
+        self.assertEqual(None, mirror.get_server())
+
     def test_finish_change_not_updating_error(self):
         basedir = get_transport(self.setup_memory()).clone('path')
         basedir.create_prefix()
@@ -183,6 +194,26 @@ exclude another regex
         self.assertEqual(('rest', 'No changes found in mirrorset.'),
             ui.outputs[-1])
 
+    def test_finish_change_no_change_recorded(self):
+        basedir = get_transport(self.setup_memory()).clone('path')
+        basedir.create_prefix()
+        ui = self.get_test_ui()
+        mirror = mirrorset.initialise(basedir, 'myname', basedir, ui)
+        basedir.create_prefix()
+        basedir.mkdir('dir1')
+        basedir.mkdir('dir2')
+        basedir.put_bytes('abc', '1234567890\n')
+        basedir.put_bytes('dir1/def', 'abcdef')
+        mirror.finish_change(dryrun=True)
+        t = basedir.clone('.lmirror/metadata/myname')
+        self.assertThat(t.get_bytes('metadata.conf'),
+            DocTestMatches("""[metadata]
+basis = 0
+latest = 0
+timestamp = ...
+updating = True
+""", ELLIPSIS))
+
     def test_finish_change_scans_content(self):
         basedir = get_transport(self.setup_memory()).clone('path')
         basedir.create_prefix()
@@ -218,6 +249,7 @@ updating = False
         basedir.put_bytes('dir1/def', 'abcdef')
         mirror.includes = ['included']
         mirror.excludes = ['dir2/', 'dir1']
+        mirror.filter_programs = []
         mirror.finish_change()
         t = basedir.clone('.lmirror/metadata/myname')
         self.assertThat(t.get_bytes('metadata.conf'), DocTestMatches("""[metadata]
